import mqtt from 'mqtt'
// import { ISubscriptionGrant, MqttClient } from 'mqtt/types/lib/client';
import type { ISubscriptionGrant, MqttClient } from 'mqtt'
// import { IConnackPacket, IDisconnectPacket, IPublishPacket } from 'mqtt-packet';
import NotifyMessage from './NotifyMessage'

export class Mqtt {
  host: string
  port: number
  username: string
  password: string
  path: string
  client?: MqttClient
  topicMap = new Map<string, Set<(message: Record<string, string>, type?: string) => void>>()

  constructor(host: string, port: number, username: string, password: string, path: string) {
    this.host = host
    this.port = port
    this.username = username
    this.password = password
    this.path = path
    this.connect()
  }

  connect() {
    this.client = mqtt.connect({
      host: this.host,
      port: this.port,
      hostname: this.host,
      username: this.username,
      password: this.password,
      path: this.path,
      protocol: location.protocol === 'http:' ? 'ws' : 'wss',
      keepalive: 15,
    })

    this.client!.once('connect', () => this.onConnect())
    this.client!.on('reconnect', () => console.debug(' mq reconnect.'))
    this.client!.on('disconnect', (packet) => console.debug(' mq disconnect.', packet))
    this.client!.on('close', () => console.debug(' mq close.'))
  }

  onConnect() {
    console.log(' mq connected.')
    this.client!.removeAllListeners('message')
    this.client!.on('message', (topic, payload, packet) =>
      this.onMessage(topic, payload.toString(), packet),
    )
    if (Array.from(this.topicMap.keys()).length > 0) {
      this.client!.subscribe(
        Array.from(this.topicMap.keys()),
        (err: Error | null, granted?: ISubscriptionGrant[]) => {
          if (err) {
            console.error('subscribe to mq error: ', err, granted)
          } else {
            // 检查 granted 是否为 undefined
            if (granted) {
              console.debug('subscribe to mq granted: ', granted)
            } else {
              console.warn('No subscription grants received.')
            }
          }
        },
      )
    }
  }

  onMessage(topic: string, payload: string, packet?: unknown) {
    try {
      console.debug(packet)
      const notify = JSON.parse(payload) as NotifyMessage
      // console.debug(notify);

      const topicCBList = this.topicMap.get(topic)
      if (topicCBList) {
        topicCBList.forEach((f) => {
          try {
            f(notify.data, notify.type)
          } catch (e) {
            console.debug(' process topic cb fun error.', e)
          }
        })
      } else {
        console.debug(" can't find subscribe callback.")
      }
    } catch (e) {
      console.error(' process message error.', e)
    }
  }

  subscribe(topic: string, callback: (message: Record<string, string>, type?: string) => void) {
    if (!this.topicMap.has(topic)) {
      this.client?.subscribe(topic, (err: Error | null, granted?: ISubscriptionGrant[]) => {
        if (err) {
          console.error('subscribe to mq error: ', err, granted)
        } else {
          // 检查 granted 是否为 undefined
          if (granted) {
            console.debug('subscribe to mq granted: ', granted)
          } else {
            console.warn(`No subscription grants received for topic: ${topic}`)
          }
        }
      })
      this.topicMap.set(topic, new Set())
    }
    this.topicMap.get(topic)?.add(callback)
  }

  unsubscribe(topic: string, callback?: (message: Record<string, string>) => void) {
    if (this.topicMap.has(topic)) {
      if (callback) {
        this.topicMap.get(topic)?.delete(callback)
      } else {
        this.topicMap.get(topic)?.clear()
      }

      if (this.topicMap.get(topic)!.size <= 0) {
        this.topicMap.delete(topic)
        this.client?.unsubscribe(topic)
      }
    }
  }
}

export class CustomConfig {
  mqttConfig?: MqttConfig

  constructor(obj: CustomConfig) {
    Object.assign(this, obj)
    if (obj.mqttConfig) {
      this.mqttConfig = new MqttConfig(obj.mqttConfig)
    }
  }
}

export class MqttConfig {
  host: string = '192.168.0.119'
  port: number = 8383
  username: string = 'admin'
  password: string = 'public'
  path: string = '/mqtt'
  use: boolean = false

  constructor(obj: MqttConfig) {
    Object.assign(this, obj)
  }

  getHost(): string {
    return this.host === '' ? document.location.hostname : this.host
  }

  getPort(): number {
    return this.port === 0 ? parseInt(document.location.port) : this.port
  }
}
